from setuptools import setup

setup(
    name='blackholepy',
    version='0.0.5',
    author='xitongsys',
    author_email='xitongsys@gmail.com',
    url='https://github.com/elephasinvestment.com/blackhole-python',
    description='blackhole python client',
    packages=['blackholepy'],
    install_requires=["crcmod"],
    entry_points={}
)
